/*
New demosystem by Nuclear / Mindlapse
Copyright (C) 2011-2013  John Tsiombikas <nuclear@member.fsf.org>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include "dsys.h"
#include "dsys_impl.h"

static int read_script(struct dsys_demo *demo, FILE *fp, const char *fname);
static char *strip_ws(char *buf);
static void dbg_print_events(struct dsys_event *ev);

static void proc_event(struct dsys_event *ev, demotime_t tm);
static void link_callback(struct dsys_event *ev, void *cls);
static void free_event(struct dsys_event *ev);

static struct dsys_event *sort_evlist(struct dsys_event *list, int num_ev);
static struct dsys_event *merge_evlists(struct dsys_event *list1, struct dsys_event *list2);


struct dsys_demo *dsys_open(const char *fname)
{
	FILE *fp;
	struct dsys_demo *demo;

	if(!(fp = fopen(fname, "r"))) {
		fprintf(stderr, "failed to open demoscript: %s: %s\n", fname, strerror(errno));
		return 0;
	}

	if(!(demo = malloc(sizeof *demo))) {
		perror("failed to allocate memory");
		fclose(fp);
		return 0;
	}
	memset(demo, 0, sizeof *demo);

	demo->src_tm = demo->start_tm = -1;

	if(read_script(demo, fp, fname) == -1) {
		free(demo);
		fclose(fp);
		return 0;
	}

	fclose(fp);
	return demo;
}

struct dsys_demo *dsys_open_stream(FILE *fp)
{
	struct dsys_demo *demo;

	if(!(demo = malloc(sizeof *demo))) {
		perror("failed to allocate memory");
		return 0;
	}
	memset(demo, 0, sizeof *demo);

	demo->src_tm = demo->start_tm = -1;

	if(read_script(demo, fp, 0) == -1) {
		free(demo);
		return 0;
	}

	return demo;
}

void dsys_close(struct dsys_demo *demo)
{
	while(demo->evlist) {
		struct dsys_event *ev = demo->evlist;
		demo->evlist = demo->evlist->next;
		free_event(ev);
	}

	free(demo);
}


#define SEP	" \t\n\r"

static int read_script(struct dsys_demo *demo, FILE *fp, const char *fname)
{
	int nline = 0;
	char buf[512], *line, *tok, *endp;
	unsigned int t0, t1;
	struct dsys_event *ev;

	if(!fname) {
		fname = "<unknown>";
	}

	demo->duration = dsys_msec_to_dtime(0);

	while(fgets(buf, sizeof buf, fp)) {
		nline++;

		line = strip_ws(buf);

		if(!line || !*line) {
			continue;
		}

		if(!(tok = strtok(line, SEP)) || (t0 = strtol(tok, &endp, 10), endp == tok)) {
			fprintf(stderr, "%s line: %d, error: expected timestamp t0\n", fname, nline);
			return -1;
		}

		if(!(tok = strtok(0, SEP))) {
			fprintf(stderr, "%s line: %d, error: expected second timestamp or event name\n", fname, nline);
			return -1;
		}

		t1 = strtol(tok, &endp, 10);
		if(endp == tok) {
			t1 = t0;
		} else {
			if(!(tok = strtok(0, SEP))) {
				fprintf(stderr, "%s line: %d, error: expected event name\n", fname, nline);
				return -1;
			}
		}

		if(!(ev = malloc(sizeof *ev))) {
			perror("read_script: failed to allocate memory for an event\n");
			return -1;
		}
		memset(ev, 0, sizeof *ev);
		ev->t0 = dsys_msec_to_dtime(t0);
		ev->t1 = dsys_msec_to_dtime(t1);

		if(!(ev->name = malloc(strlen(tok) + 1))) {
			free(ev);
			fprintf(stderr, "read_script: failed to allocate memory for the event name: %s\n", tok);
			return -1;
		}
		strcpy(ev->name, tok);

		ev->eval = t0 == t1 ? dsys_eval_step : dsys_eval_lerp;

		ev->next = demo->evlist;
		ev->prev = 0;
		if(demo->evlist) {
			demo->evlist->prev = ev;
		}
		demo->evlist = ev;
		demo->num_ev++;

		if(ev->t1 > demo->duration) {
			demo->duration = ev->t1;
		}
	}

	demo->evlist = sort_evlist(demo->evlist, demo->num_ev);

	/*dbg_print_events(demo->evlist);*/

	return 0;
}

static char *strip_ws(char *buf)
{
	char *ptr;

	while(isspace(*buf)) {
		buf++;
	}

	ptr = buf;
	while(*ptr) {
		if(*ptr == '\n' || *ptr == '\r' || *ptr == '#') {
			*ptr = 0;
			break;
		}
		ptr++;
	}

	return buf;
}

static void dbg_print_events(struct dsys_event *ev)
{
	int i;

	for(i=0; ev; i++) {
		printf("%02d - %s (%f -> %f) [%s]\n", i, ev->eval == dsys_eval_step ? "step" : "lerp",
				ev->t0, ev->t1, ev->name);
		ev = ev->next;
	}
}

void dsys_update(struct dsys_demo *demo, demotime_t tm)
{
	struct dsys_event *ev;

	demo->src_tm = tm;

	if(demo->start_tm == -1) {
		dsys_start(demo);
	}

	if(!demo->running) {
		return;	/* nothing changes */
	}

	demo->tm = tm - demo->start_tm - demo->stoppage_tm;

	if(demo->tm < 0) {
		demo->tm = 0;
	}
	if(demo->tm > demo->duration) {
		demo->tm = demo->duration;
	}

	while(demo->active && demo->active->t1 <= demo->tm) {
		proc_event(demo->active, demo->tm);
		demo->active = demo->active->next;
	}

	ev = demo->active;
	while(ev && ev->t0 <= demo->tm) {
		proc_event(ev, demo->tm);
		ev = ev->next;
	}
	demo->nextev = ev;


	if(demo->tm >= demo->duration) {
		dsys_stop(demo);
	}
}

static void proc_event(struct dsys_event *ev, demotime_t tm)
{
	float val = ev->eval(ev, tm);

	if(ev->val != val) {
		struct callback *cb = ev->cblist;

		ev->val = val;

		while(cb) {
			cb->func(ev, cb->cls);
			cb = cb->next;
		}
	}
}

void dsys_start(struct dsys_demo *demo)
{
	if(demo->running) {
		return;
	}

	if(demo->start_tm == -1) {
		demo->start_tm = demo->src_tm;
		demo->nextev = demo->active = demo->evlist;
	} else {
		demo->stoppage_tm += demo->src_tm - demo->stop_tm;
	}

	demo->running = 1;
}

void dsys_stop(struct dsys_demo *demo)
{
	if(!demo->running) {
		return;
	}

	demo->stop_tm = demo->src_tm;
	demo->running = 0;
}

int dsys_is_running(struct dsys_demo *demo)
{
	return demo->running;
}


demotime_t dsys_duration(struct dsys_demo *demo)
{
	return demo->duration;
}

demotime_t dsys_time(struct dsys_demo *demo)
{
	return demo->tm;
}

float dsys_progress(struct dsys_demo *demo)
{
	return demo->tm / demo->duration;
}

/* seek without continuity */
void dsys_seek(struct dsys_demo *demo, demotime_t tm)
{
	struct dsys_event *ev;

	if(tm < 0) {
		tm = 0;
	}
	if(tm > demo->duration) {
		tm = demo->duration;
	}

	if(tm < demo->tm) {
		/* on backwards seek, invalidate the sliding window */
		demo->nextev = demo->active = demo->evlist;
	}

	demo->start_tm = demo->src_tm - tm;
	demo->stoppage_tm = 0;
	demo->stop_tm = demo->src_tm;
	demo->tm = tm;

	/* recalculate events */
	ev = demo->evlist;
	while(ev) {
		proc_event(ev, tm);
		ev = ev->next;
	}
}

void dsys_seek_norm(struct dsys_demo *demo, float t)
{
	dsys_seek(demo, t * demo->duration);
}

/* seek by accelerating time */
void dsys_warp(struct dsys_demo *demo, demotime_t tm)
{
	fprintf(stderr, "dsys_warp not implemented yet\n");
}

void dsys_warp_norm(struct dsys_demo *demo, float t)
{
	dsys_warp(demo, t * demo->duration);
}


/* events */
struct dsys_event *dsys_event(struct dsys_demo *demo, const char *name)
{
	struct dsys_event *iter = demo->evlist;

	while(iter) {
		if(strcmp(iter->name, name) == 0) {
			return iter;
		}
		iter = iter->next;
	}
	return 0;
}

void dsys_set_event_eval(struct dsys_event *ev, float (*eval)(struct dsys_event*, demotime_t))
{
	ev->eval = eval;
}

enum dsys_evtype dsys_event_type(struct dsys_event *ev)
{
	return ev->type;
}

float dsys_event_value(struct dsys_event *ev)
{
	return ev->val;
}

int dsys_event_callback(struct dsys_event *ev, void (*func)(struct dsys_event*, void*), void *cls)
{
	struct callback *cb;

	if(!(cb = malloc(sizeof *cb))) {
		perror("failed to allocate memory");
		return -1;
	}
	cb->func = func;
	cb->cls = cls;
	cb->next = ev->cblist;
	ev->cblist = cb;
	return 0;
}

int dsys_event_link(struct dsys_event *ev, float *link)
{
	return dsys_event_callback(ev, link_callback, link);
}

static void link_callback(struct dsys_event *ev, void *cls)
{
	*(float*)cls = ev->val;
}


/* time conversion */
demotime_t dsys_sec_to_dtime(float sec)
{
	return sec;
}

demotime_t dsys_msec_to_dtime(unsigned long msec)
{
	return (demotime_t)msec / 1000.0;
}

float dsys_dtime_to_sec(demotime_t tm)
{
	return tm;
}

unsigned long dsys_dtime_to_msec(demotime_t tm)
{
	return (unsigned long)(tm * 1000.0);
}


float dsys_eval_step(struct dsys_event *ev, demotime_t t)
{
	return t >= ev->t1 ? 1.0 : 0.0;
}

#define CLAMP(x, low, high)	((x) < (low) ? (low) : ((x) > (high) ? (high) : (x)))

float dsys_eval_lerp(struct dsys_event *ev, demotime_t t)
{
	float res = (t - ev->t0) / (ev->t1 - ev->t0);
	return CLAMP(res, 0.0, 1.0);
}

float dsys_eval_sigmoid(struct dsys_event *ev, demotime_t t)
{
	t = dsys_eval_lerp(ev, t);
	return 1.0 - (cos(t * M_PI) * 0.5 + 0.5);
}

static void free_event(struct dsys_event *ev)
{
	while(ev->cblist) {
		struct callback *cb = ev->cblist;
		ev->cblist = ev->cblist->next;
		free(cb);
	}
}

static struct dsys_event *sort_evlist(struct dsys_event *list, int num_ev)
{
	int i, num_left, num_right;
	struct dsys_event *left, *right, *node = list;

	if(num_ev < 2) {
		return list;
	}

	num_left = num_ev / 2;
	num_right = num_ev - num_left;

	for(i=0; i<num_ev/2; i++) {
		node = node->next;
	}

	if(node->prev) {
		node->prev->next = 0;
		node->prev = 0;
	}

	left = sort_evlist(list, num_left);
	right = sort_evlist(node, num_right);

	return merge_evlists(left, right);
}

static struct dsys_event *merge_evlists(struct dsys_event *list1, struct dsys_event *list2)
{
	struct dsys_event *head, *tail, *node;

	if(!list1) {
		return list2;
	}
	if(!list2) {
		return list1;
	}

	head = tail = 0;

	while(list1 && list2) {
		if(list1->t0 < list2->t0) {
			node = list1;
			list1 = list1->next;
		} else {
			node = list2;
			list2 = list2->next;
		}

		node->next = 0;
		node->prev = tail;

		if(!head) {
			head = node;
		} else {
			tail->next = node;
		}
		tail = node;
	}

	if(list1) {
		tail->next = list1;
		list1->prev = tail;
	} else if(list2) {
		tail->next = list2;
		list2->prev = tail;
	}

	return head;
}
