/*
Printblobs - halftoning display hack
Copyright (C) 2013  John Tsiombikas <nuclear@member.fsf.org>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <vector>
#include <math.h>
#include "opengl.h"
#include "mballs.h"
#include "metasurf.h"
#include "dsys.h"
#include "udg.h"

struct MetaBall {
	float pos[3];
	float orbit;
	float energy;
	float phase_offs;
};

#define VOL_SZ			2
#define MBALL_GRID_SZ	50

static void update(float sec);
static float calc_field(float x, float y, float z);
static float eval(float x, float y, float z);
static void vertex(float x, float y, float z);
static void normal(float x, float y, float z);

static float grid[MBALL_GRID_SZ][MBALL_GRID_SZ][MBALL_GRID_SZ];
static std::vector<MetaBall> balls;
static struct metasurface *msurf;
static float floor_height = -0.95;

static struct dsys_event *evrise;

bool mball_init()
{
	static const float bbmin = -VOL_SZ / 2.0;
	static const float bbmax = VOL_SZ / 2.0;

	if(!(msurf = msurf_create())) {
		return false;
	}
	msurf_threshold(msurf, 10);
	msurf_resolution(msurf, MBALL_GRID_SZ, MBALL_GRID_SZ, MBALL_GRID_SZ);
	msurf_bounds(msurf, bbmin, bbmin, bbmin, bbmax, bbmax, bbmax);
	msurf_eval_func(msurf, eval);
	msurf_vertex_func(msurf, vertex);
	//msurf_normal_func(msurf, normal);

	for(int i=0; i<10; i++) {
		MetaBall mb;
		mb.orbit = 0.25 * rand() / (float)RAND_MAX + 0.35;
		mb.energy = 0.1 * rand() / (float)RAND_MAX + 0.15;
		mb.phase_offs = rand() / (float)RAND_MAX * M_PI * 2.0;
		balls.push_back(mb);
	}

	evrise = dsys_event(demo, "mball_rise");
	dsys_set_event_eval(evrise, dsys_eval_sigmoid);

	return true;
}

void mball_render(float sec)
{
	update(sec);

	const float blue[] = {0.4, 0.45, 1.0, 1};
	const float dark_red[] = {0.6, 0.2, 0.1, 1};
	const float white[] = {1, 1, 1, 1};
	const float black[] = {0, 0, 0, 1};

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glScalef(4.0, 4.0, 4.0);

	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, blue);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, white);
	glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 80.0);

	glBegin(GL_TRIANGLES);
	msurf_polygonize(msurf);
	glEnd();

	// floor
	glBegin(GL_QUADS);
	glNormal3f(0, 1, 0);
	glVertex3f(-5, -1, 5);
	glVertex3f(5, -1, 5);
	glVertex3f(5, -1, -5);
	glVertex3f(-5, -1, -5);
	glEnd();

	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, dark_red);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, black);
	glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 80.0);

	// box
	glPushMatrix();
	glTranslatef(0, -1.7, 1);
	glScalef(1.05, 1, 0.05);
	glutSolidCube(2.0);
	glPopMatrix();

	glPushMatrix();
	glTranslatef(0, -1.7, -1);
	glScalef(1.05, 1, 0.05);
	glutSolidCube(2.0);
	glPopMatrix();

	glPushMatrix();
	glTranslatef(1, -1.7, 0);
	glScalef(0.05, 1, 1);
	glutSolidCube(2.0);
	glPopMatrix();

	glPushMatrix();
	glTranslatef(-1, -1.7, 0);
	glScalef(0.05, 1, 1);
	glutSolidCube(2.0);
	glPopMatrix();

	glPopMatrix();
}


static void update(float sec)
{
	float trise = dsys_event_value(evrise);

	for(size_t i=0; i<balls.size(); i++) {
		float t = sec + balls[i].phase_offs;
		balls[i].pos[0] = cos(t * 1.8) * balls[i].orbit;
		balls[i].pos[2] = sin(t * 1.2) * balls[i].orbit;
		balls[i].pos[1] = (sin(t) + cos(t * 2.0) / 2.0 + sin(t * 3.0) / 3.0) * 0.45 - (2.0 - trise * 2.0);
	}

	for(int i=0; i<MBALL_GRID_SZ; i++) {
		float x = (float)i / (float)MBALL_GRID_SZ * 2.0 - 1.0;
		for(int j=0; j<MBALL_GRID_SZ; j++) {
			float y = (float)j / (float)MBALL_GRID_SZ * 2.0 - 1.0;
			for(int k=0; k<MBALL_GRID_SZ; k++) {
				float z = (float)k / (float)MBALL_GRID_SZ * 2.0 - 1.0;
				grid[i][j][k] = calc_field(x, y, z);
			}
		}
	}
}

static float calc_field(float x, float y, float z)
{
	float sum = 0.0f;
	for(size_t i=0; i<balls.size(); i++) {
		float dx = balls[i].pos[0] - x;
		float dy = balls[i].pos[1] - y;
		float dz = balls[i].pos[2] - z;
		float dist_sq = dx * dx + dy * dy + dz * dz;
		if(dist_sq > 1e-6) {
			sum += balls[i].energy / dist_sq;
		} else {
			sum += 1000.0;
		}
	}

	// floor
	float height = y - floor_height;
	if(height > 1e-6) {
		sum += 1.0 / height;
	} else {
		sum += 1000.0;
	}

	return sum;
}

static inline int clamp(int x, int a, int b)
{
	return x < a ? a : (x > b ? b : x);
}

static float eval(float x, float y, float z)
{
	int cell_x = clamp((int)((x / VOL_SZ + 0.5) * MBALL_GRID_SZ), 0, MBALL_GRID_SZ - 1);
	int cell_y = clamp((int)((y / VOL_SZ + 0.5) * MBALL_GRID_SZ), 0, MBALL_GRID_SZ - 1);
	int cell_z = clamp((int)((z / VOL_SZ + 0.5) * MBALL_GRID_SZ), 0, MBALL_GRID_SZ - 1);

	return grid[cell_x][cell_y][cell_z];
}

static void vertex(float x, float y, float z)
{
	float delta = (float)VOL_SZ / (float)MBALL_GRID_SZ;

	float dfdx = calc_field(x - delta, y, z) - calc_field(x + delta, y, z);
	float dfdy = calc_field(x, y - delta, z) - calc_field(x, y + delta, z);
	float dfdz = calc_field(x, y, z - delta) - calc_field(x, y, z + delta);

	float len = sqrt(dfdx * dfdx + dfdy * dfdy + dfdz * dfdz);

	glNormal3f(dfdx / len, dfdy / len, dfdz / len);
	glVertex3f(x, y, z);
}

static void normal(float x, float y, float z)
{
	glNormal3f(x, y, z);
}
