/*
Printblobs - halftoning display hack
Copyright (C) 2013  John Tsiombikas <nuclear@member.fsf.org>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include "opengl.h"
#include "udg.h"
#include "sdr.h"
#include "dither_matrix.h"
#include "mballs.h"
#include "dsys.h"
#include "post.h"

#define DITHER_SZ		8
#define DITHER_LEVELS	16

#if DITHER_SZ == 4
#define dither_matrix	dither_matrix4
#elif DITHER_SZ == 8
#define dither_matrix	halftone_matrix8
#else
#error "invalid dither size"
#endif

struct render_target {
	unsigned int fbo;
	unsigned int color_tex, depth_buf;
};

bool init();
void cleanup();
void disp();
void idle();
void reshape(int x, int y);
void keyb(unsigned char key, int x, int y);
void mouse(int bn, int state, int x, int y);
void motion(int x, int y);
struct render_target *create_rtarg(int xsz, int ysz);
void destroy_rtarg(struct render_target *rt);
bool parse_args(int argc, char **argv);

int xsz, ysz;
float cam_theta, cam_phi = 25, cam_dist = 11;
unsigned int dither_tex;
struct render_target *rtarg;
unsigned int post_prog, phong_prog;

int opt_blocky = true, opt_regular_render;
bool opt_autorot = true;

struct dsys_demo *demo;
struct dsys_event *evfadein;

int main(int argc, char **argv)
{
	glutInitWindowSize(1024, 768);
	glutInit(&argc, argv);

	if(!parse_args(argc, argv)) {
		return 1;
	}

	glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);
	glutCreateWindow("DBF UDG compo entry by Nuclear");

	glutDisplayFunc(disp);
	glutIdleFunc(idle);
	glutReshapeFunc(reshape);
	glutKeyboardFunc(keyb);
	glutMouseFunc(mouse);
	glutMotionFunc(motion);

	glewInit();

	if(!init()) {
		return 1;
	}

	glutMainLoop();
	return 0;
}

bool init()
{
	if(!(demo = dsys_open("demoscript"))) {
		return false;
	}
	evfadein = dsys_event(demo, "fade_in");
	dsys_set_event_eval(evfadein, dsys_eval_sigmoid);

	// dump the tile image
	FILE *fp = fopen("udg.ppm", "wb");
	if(fp) {
		fprintf(fp, "P6\n%d %d\n255\n", DITHER_SZ, DITHER_SZ * DITHER_LEVELS);
	}

	unsigned char *img = new unsigned char[DITHER_SZ * DITHER_SZ * DITHER_LEVELS];
	unsigned char *ptr = img;

	for(int i=0; i<DITHER_LEVELS; i++) {
		float val = (float)i / (float)(DITHER_LEVELS - 1);
		for(int y=0; y<DITHER_SZ; y++) {
			for(int x=0; x<DITHER_SZ; x++) {
				/* (1 + M) / (1 + MxN) */
				float thres = (1.0 + dither_matrix[x][y]) / (1.0 + DITHER_SZ * DITHER_SZ);
				*ptr++ = val >= thres ? 255 : 0;

				if(fp) {
					int r = ptr[-1] ? 246 : 10;
					int g = ptr[-1] ? 198 : 72;
					int b = ptr[-1] ? 141 : 85;
					fputc(r, fp);
					fputc(g, fp);
					fputc(b, fp);
				}
			}
		}
	}

	if(fp) {
		fclose(fp);
	}

	if(!(phong_prog = create_program_load("sdr/phong.v.glsl", "sdr/phong.p.glsl"))) {
		return false;
	}

	if(!(post_prog = create_program_load("sdr/dither.v.glsl", "sdr/dither.p.glsl"))) {
		return false;
	}

	glGenTextures(1, &dither_tex);
	glBindTexture(GL_TEXTURE_2D, dither_tex);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_LUMINANCE, DITHER_SZ, DITHER_SZ * DITHER_LEVELS, 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, img);

	if(!mball_init()) {
		return false;
	}

	glEnable(GL_CULL_FACE);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glEnable(GL_LIGHT1);
	glEnable(GL_NORMALIZE);


	reshape(glutGet(GLUT_WINDOW_WIDTH), glutGet(GLUT_WINDOW_HEIGHT));
	return true;
}

void draw_backdrop()
{
	glPushAttrib(GL_ENABLE_BIT);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();

	glBegin(GL_QUADS);
	glColor3f(0, 0, 0);
	glVertex2f(-1, -1);
	glVertex2f(1, -1);
	glColor3f(1, 1, 1);
	glVertex2f(1, 1);
	glVertex2f(-1, 1);
	glEnd();

	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

	glPopAttrib();
}

void disp()
{
	float ldir[] = {-1, 1, 2, 0};
	float ldir2[] = {0.0, 0.35, -0.9, 0};

	float lcol[] = {1, 1, 1, 1};
	float lcol2[] = {0.35, 0.3, 0.15, 1};

	dsys_update(demo, dsys_msec_to_dtime(glutGet(GLUT_ELAPSED_TIME)));

	float sec = dsys_dtime_to_sec(dsys_time(demo));
	float auto_angle = sec * 10.0;

	int xres, yres;
	if(!opt_blocky) {
		xres = xsz;
		yres = ysz;
	} else {
		xres = xsz / DITHER_SZ;
		yres = ysz / DITHER_SZ;
	}

	if(!rtarg) {
		printf("(re)creating render target (%dx%d)\n", xres, yres);
		if(!(rtarg = create_rtarg(xres, yres))) {
			exit(0);
		}
	}

	if(!opt_regular_render) {
		glBindFramebufferEXT(GL_FRAMEBUFFER, rtarg->fbo);
	}
	glViewport(0, 0, xres, yres);

	glClearColor(1, 1, 1, 1);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	draw_backdrop();

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	glTranslatef(0, 0.8, -cam_dist);
	glRotatef(cam_phi, 1, 0, 0);
	glRotatef(opt_autorot ? auto_angle : cam_theta, 0, 1, 0);

	glLightfv(GL_LIGHT0, GL_POSITION, ldir);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, lcol);

	glLightfv(GL_LIGHT1, GL_POSITION, ldir2);
	glLightfv(GL_LIGHT1, GL_DIFFUSE, lcol2);

	const float blue[] = {0.4, 0.45, 1.0, 1};
	const float white[] = {1, 1, 1, 1};
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, blue);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, white);
	glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 80.0);

	bind_program(phong_prog);
	mball_render(sec);
	bind_program(0);

	float tfadein = evfadein ? dsys_event_value(evfadein) : 1.0;
	/*if(tfadein < 1.0) {
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		overlay(0, 0, 0, 1.0 - tfadein);
		glDisable(GL_BLEND);
	}*/


	if(!opt_regular_render) {
		glBindFramebufferEXT(GL_FRAMEBUFFER, 0);
		glViewport(0, 0, xsz, ysz);

		glClear(GL_COLOR_BUFFER_BIT);

		bind_program(post_prog);
		set_uniform_int(post_prog, "framebuf", 0);
		set_uniform_int(post_prog, "dither_tex", 1);
		set_uniform_int(post_prog, "dither_levels", DITHER_LEVELS);
		set_uniform_int(post_prog, "dither_size", DITHER_SZ);
		set_uniform_float(post_prog, "tfadein", tfadein);

		glActiveTextureARB(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D, rtarg->color_tex);
		glEnable(GL_TEXTURE_2D);
		glActiveTextureARB(GL_TEXTURE1);
		glBindTexture(GL_TEXTURE_2D, dither_tex);
		glEnable(GL_TEXTURE_2D);

		overlay(1, 1, 1, 1);

		glActiveTextureARB(GL_TEXTURE1);
		glDisable(GL_TEXTURE_2D);
		glActiveTextureARB(GL_TEXTURE0);
		glDisable(GL_TEXTURE_2D);

		bind_program(0);
	}

	glutSwapBuffers();
	assert(glGetError() == GL_NO_ERROR);
}

void idle()
{
	glutPostRedisplay();
}

void reshape(int x, int y)
{
	glViewport(0, 0, x, y);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(45.0, (float)x / (float)y, 0.5, 500.0);

	if(x != xsz || y != ysz) {
		destroy_rtarg(rtarg);
		rtarg = 0;
		xsz = x;
		ysz = y;
	}
}

void keyb(unsigned char key, int x, int y)
{
	switch(key) {
	case 27:
		exit(0);

	case 'a':
		opt_autorot = !opt_autorot;
		break;

	case 'f':
		{
			static bool fullscreen;
			static int orig_x, orig_y;

			fullscreen = !fullscreen;
			if(fullscreen) {
				orig_x = xsz;
				orig_y = ysz;
				glutFullScreen();
			} else {
				glutReshapeWindow(orig_x, orig_y);
			}
		}
		break;

	case 'r':
		opt_regular_render = !opt_regular_render;
		break;

	case 'b':
		opt_blocky = !opt_blocky;
		if(rtarg) {
			destroy_rtarg(rtarg);
			rtarg = 0;
		}
		break;

	case ' ':
		if(dsys_is_running(demo)) {
			dsys_stop(demo);
		} else {
			dsys_start(demo);
		}
		break;
	}
}

bool bnstate[16];
int prev_x, prev_y;

void mouse(int bn, int state, int x, int y)
{
	int idx = bn - GLUT_LEFT_BUTTON;

	if(idx < (int)(sizeof bnstate / sizeof *bnstate)) {
		bnstate[idx] = state == GLUT_DOWN;
	}
	prev_x = x;
	prev_y = y;
}

void motion(int x, int y)
{
	int dx = x - prev_x;
	int dy = y - prev_y;
	prev_x = x;
	prev_y = y;

	if(bnstate[0]) {
		cam_theta = fmod(cam_theta + dx * 0.5, 360.0);
		cam_phi += dy * 0.5;
		if(cam_phi < -90) {
			cam_phi = -90;
		}
		if(cam_phi > 90) {
			cam_phi = 90;
		}
	}
	if(bnstate[2]) {
		cam_dist += dy * 0.1;
		if(cam_dist < 0) {
			cam_dist = 0;
		}
	}
}

struct render_target *create_rtarg(int xsz, int ysz)
{
	struct render_target *rt = new render_target;

	glGenFramebuffersEXT(1, &rt->fbo);
	glBindFramebufferEXT(GL_FRAMEBUFFER, rt->fbo);

	// create the render target texture
	glGenTextures(1, &rt->color_tex);
	glBindTexture(GL_TEXTURE_2D, rt->color_tex);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, xsz, ysz, 0, GL_RGB, GL_UNSIGNED_BYTE, 0);

	glFramebufferTexture2DEXT(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, rt->color_tex, 0);

	// create depth buffer
	glGenRenderbuffersEXT(1, &rt->depth_buf);
	glBindRenderbufferEXT(GL_RENDERBUFFER, rt->depth_buf);
	glRenderbufferStorageEXT(GL_RENDERBUFFER, GL_DEPTH_COMPONENT, xsz, ysz);

	glFramebufferRenderbufferEXT(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, rt->depth_buf);

	if(glCheckFramebufferStatusEXT(GL_FRAMEBUFFER) != GL_FRAMEBUFFER_COMPLETE) {
		fprintf(stderr, "incomplete fbo\n");
		return 0;
	}

	glBindFramebufferEXT(GL_FRAMEBUFFER, 0);
	return rt;
}

void destroy_rtarg(struct render_target *rt)
{
	if(!rt) {
		return;
	}
	glDeleteFramebuffersEXT(1, &rt->fbo);
	glDeleteTextures(1, &rt->color_tex);
	glDeleteRenderbuffersEXT(1, &rt->depth_buf);
	delete rt;
}

bool parse_args(int argc, char **argv)
{
	for(int i=1; i<argc; i++) {
		if(strcmp(argv[i], "-noblocky") == 0) {
			opt_blocky = false;
		} else if(strcmp(argv[i], "-blocky") == 0) {
			opt_blocky = true;
		} else if(strcmp(argv[i], "-nodither") == 0) {
			opt_regular_render = true;
		} else {
			fprintf(stderr, "invalid argument: %s\n", argv[i]);
			return false;
		}
	}
	return true;
}
